
# These should use the "quick" option, I think...
proc CmdToright {} {
  global glob
  NewPwd right $glob(left,pwd)
  UpdateWindow_ right 1
}

proc CmdToleft {} {
  global glob
  NewPwd left $glob(right,pwd)
  UpdateWindow_ left 1
}

proc CmdSwapWindows {} {
  global glob
  set tmpright $glob(left,pwd)
  set glob(forceupdate) 1
  CmdToleft
  NewPwd right $tmpright
  UpdateWindow right
  set glob(forceupdate) 0
}

proc BatchReceiveFTP { inst } {
  global glob
  if {[IsFTP $glob($inst,pwd)]} {
    PopInfo [_ "You can only issue a receive request to a non-FTP directory"]
    return
  }
  set olddir $glob([Opposite $inst],pwd)
  foreach itemblock $glob(batchlist) {
    set item [lindex $itemblock 0]
    set r [regexp {ftp://([^/]*)(.*)} $item match ftpI elem]
    if {!$r} {
      PopWarn [_ "Can't parse %s as FTP file" $item]
    } else {
      NewPwd [Opposite $inst] ftp://$ftpI/
      set r [Try { FTP_GetFile $ftpI "$elem" \
		       "$glob($inst,pwd)/[file tail $elem]" \
		       [lindex $itemblock 1] 0 } "" 1]
    }
  }
  set glob(batchlist) {}
  set glob(forceupdate) 1
  NewPwd [Opposite $inst] $olddir
  UpdateWindow both
  set glob(forceupdate) 0
}


proc CmdCopy {{resume 0}} {
  global glob
  if { $glob(left,pwd) == $glob(right,pwd) } {
    tk_dialog .apop [_ "Oops That doesn't work"] \
       [_ "Please set different source and destination directories!"] "" 0 [_ "OK"] 
    return {}
  }    
  CmdCopy_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd) left $resume ""
  CmdCopy_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) right $resume ""
  UpdateWindow both
  set glob(forceupdate) 0
}
proc CmdHardlnk {{resume 0}} {
  global glob
  global config
  if { $config(cmd,hardlink) == "" } {
    tk_dialog .apop [_ "Sorry"] \
	[_ "The hard link command is not configured (see Configuration menu)."]\
	"" 0 [_ "OK"]
    return {}
  }
  if { $glob(left,pwd) == $glob(right,pwd) } {
    tk_dialog .apop [_ "Oops That doesn't work"] \
      [_ "Please set different source and destination directories!"] "" 0 [_ "OK"] 
    return {}
  }
  if { [lindex $config(cmd,hardlink) 1] == "r" } {
    set r left
    set l right
  } else {
    set r right
    set l left
  }
  CmdCopy_ $glob(listbox,left).file glob(left,filelist) \
      $glob($l,pwd) $glob($r,pwd) left $resume [lindex $config(cmd,hardlink) 0]
  CmdCopy_ $glob(listbox,right).file glob(right,filelist) \
      $glob($r,pwd) $glob($l,pwd) right $resume [lindex $config(cmd,hardlink) 0]
  UpdateWindow both
  set glob(forceupdate) 0
}

proc CmdCopy_ { listb_name filelist_var frompwd topwd inst resume option} {
  # For speed, (hopefully) we'll call by reference...
  global config glob
  upvar $filelist_var filelist
  set cpcmd $option
  if { $cpcmd == "" } {
    set cpcmd $config(cmd,cp)
  }

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Copy"]]} return
    set elem [lindex $filelist $sel]
    Log [_ "%s switching on %s" CmdCopy_ [lindex $elem 2]]
    switch [lindex $elem 2] {
      fld -
      fd  { 
        if {[IsFTP $topwd]} {
	  Log [_ "IsFTP source and destination(1)"]
          CantDoThat
          return
        }
        set olddir $glob($inst,pwd)
        set r [Try { 
	  CopyFromFTPRecursive \
	      "$frompwd/[lindex $elem 1]" "$topwd" \
	      $inst $resume \
	} "" 1]
        NewPwd $inst $olddir
        UpdateWindow $inst
    }
      fl  -
      fn  {
	if { $option != "" } {
	  Log [_ "IsFTP source and options %s" $option]
          CantDoThat
          return
        }	      
        if {[IsFTP $topwd]} {
	  Log [_ "IsFTP source and destination(2)"]
          CantDoThat
          return
        }
        set r [regexp {ftp://([^/]*)(.*)} $frompwd match ftpI directory]
        if {$r == 0} { 
          PopError [_ "Can't parse %s as ftp URL" $frompwd] 
        } else {
          set size [lindex $elem 3]
          if {[lindex $elem 2] == "fl"} {set size -1}
          if {$glob(async)} {
            catch {exec $glob(lib_fr)/frftp $glob(lib_fr) \
		       $config(ftp,timeout) 1 $glob(ftp,$ftpI,host) \
                       $glob(ftp,$ftpI,user) $glob(ftp,$ftpI,passwd) \
		       $ftpI "$directory/[lindex $elem 1]" \
                       "$topwd/[lindex $elem 1]" $size $resume &} out
          } else {
            set r [Try { FTP_GetFile $ftpI \
			     "$directory/[lindex $elem 1]" \
			     "$topwd/[lindex $elem 1]" \
			     $size $resume } "" 1]
          }
        }
    }
      n -
      l   { 
        Log [_ "Copying (1) %s to %s" \
		 "$frompwd/[lindex $elem 1] $topwd"]
        set r [regexp {ftp://([^/]*)(.*)} $topwd match ftpI directory]
        if {$r} {
	  if { $option != "" } {
	        Log [_ "ftp: and options %s" $option]
		CantDoThat
		return
	    }	      
          set glob(forceupdate) 1
          Try { FTP_PutFile $ftpI $frompwd/[lindex $elem 1] \
		    $directory/[lindex $elem 1] [lindex $elem 3] } "" 1
        } else {
          if {[CheckWhoOwns $topwd/[lindex $elem 1] overwrite]} {
	    # Use "" on the try script & not {} so the command is expaned and not
	    # taken as one parameter (hard to find 'cp -f')
	    if {$cpcmd != ""} {
	      Try "exec $cpcmd  \
                [FixFileName $frompwd/[lindex $elem 1] 2 ] \
                [FixFileName $topwd ]" "" 1 $glob(async)
	    } else {
	      Try "file copy -force  \
                [FixFileName $frompwd/[lindex $elem 1] 2 ] \
                [FixFileName $topwd ]" "" 1 
	    }
          }
        }
       }

      d -
      ld   { 
        if {[IsFTP $topwd]} {
	  if { $option != "" } {
	        Log [_ "IsFTP and options %s" $option]
		CantDoThat
		return
	    }	      
          set olddir $glob($inst,pwd)
          set r [Try { CopyToFTPRecursive \
			   "$frompwd/[lindex $elem 1]" "$topwd" $inst } "" 1]
          NewPwd $inst $olddir
          set glob(forceupdate) 1
	  UpdateWindow $inst
	  set glob(forceupdate) 0
        } else {
          Log [_ "Copying %s to %s" \
		  "$config(cmd,cp) $option $frompwd/[lindex $elem 1]" $topwd]
          if {[CheckWhoOwns $topwd/[lindex $elem 1] overwrite]} {
	    if { $cpcmd == "" } {
	      set cpcmd $config(cmd,dircp)
	    }
	    if { $cpcmd == "" } {
	      Try "file copy \
                [FixFileName $frompwd/[lindex $elem 1] ] \
                [FixFileName $topwd/[lindex $elem 1] ]" "" 1 
	    } else {
	      Try "exec $cpcmd \
                [FixFileName $frompwd/[lindex $elem 1] ] \
                [FixFileName $topwd/[lindex $elem 1] ]" "" 1 $glob(async)
	    }
          }
        }
    }
      default {
	  Log [_  "file type %s ??" [lindex $elem 2]]
	  CantDoThat
      }
    }
  }
}

proc CmdCopyAs {} {
  global glob
  CmdCopyAs_ $glob(listbox,left).file glob(left,filelist) $glob(left,pwd) $glob(right,pwd)
  CmdCopyAs_ $glob(listbox,right).file glob(right,filelist) $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
}

proc CmdCopyAs_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  global config glob
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_  "CopyAs"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      fl  -
      fn  {
        set destfile [EntryDialog "" [_  "Copy As..."] \
			  [_  "Please enter new name for destination file"] \
			  $topwd/[lindex $elem 1] question]
        if {$destfile != "" } {
          if {[IsFTP $destfile]} {
            CantDoThat
            return
          }
          set r [regexp {ftp://([^/]*)(.*)} $frompwd match ftpI directory]
          if {$r == 0} { 
            PopError [_ "Can't parse %s as ftp URL" $frompwd] 
          } else {
            set size [lindex $elem 3]
            if {[lindex $elem 2] == "fl"} {set size -1}
            if {$glob(async)} {
              catch {exec $glob(lib_fr)/frftp $glob(lib_fr)\
			 $config(ftp,timeout) 1 $glob(ftp,$ftpI,host) \
                         $glob(ftp,$ftpI,user) $glob(ftp,$ftpI,passwd)\
			 $ftpI "$directory/[lindex $elem 1]" \
                         "$destfile" $size 0 &} out
            } else {
              set r [Try { FTP_GetFile $ftpI \
			       "$directory/[lindex $elem 1]"\
			       "$destfile" $size 0 } "" 1]
            }
          }
        }
      }
      l   { 
        set destfile [EntryDialog  "" [_  "Copy As..."] \
			  [_  "Please enter new name for destination file"] \
			  $topwd/[lindex $elem 1] question]
        if {[IsFTP $destfile]} {
          CantDoThat
          return
        }
        if {$destfile != "" } {
          Log [_ "Copying %s to %s" \
		  "$config(cmd,cp) $frompwd/[lindex $elem 1]" $destfile]
	  if { $config(cmd,cp) == "" } {
	    Try "file copy \
                [FixFileName $frompwd/[lindex $elem 1]] \
                [FixFileName $destfile]" "" 1  $glob(async)
	  } else {
	    Try "exec $config(cmd,cp) \
                [FixFileName $frompwd/[lindex $elem 1]] \
                [FixFileName $destfile]" "" 1  $glob(async)
	  }
        }
      }
      n   { 
        set destfile [EntryDialog "" [_  "Copy As..."] \
			  [_  "Please enter new name for destination file"] \
			  $topwd/[lindex $elem 1] question]
        if {[IsFTP $destfile]} {
          CantDoThat
          return
        }
        if {$destfile != "" } {
          Log [_ "Copying %s to %s" \
		  "$config(cmd,cp) $frompwd/[lindex $elem 1]" $destfile]
	  if { $config(cmd,cp) == "" } {
          Try "file copy \
              [FixFileName $frompwd/[lindex $elem 1]]\
              [FixFileName $destfile] " "" 1  $glob(async)
	  } else {
          Try "exec $config(cmd,cp) \
              [FixFileName $frompwd/[lindex $elem 1]]\
              [FixFileName $destfile] " "" 1  $glob(async)
	  }
#          Try { file copy -force -- $frompwd/[lindex $elem 1] \
#	      $destfile } "" 1  $glob(async)
        }
      }
      d   -
      ld  { 
        set destdir [EntryDialog  "" [_  "Copy As..."] \
			 [_  "Please enter new name for directory after copy"] \
			 $topwd/[lindex $elem 1] question]
        if {[IsFTP $destdir]} {
          CantDoThat
          return
        }
        if {$destdir != "" } {
          Log [_  "Copying %s to %s" \
		  "$config(cmd,cp) $frompwd/[lindex $elem 1]" $destdir]
	  if { $config(cmd,cp) == "" } {
	  } else {
          Try "exec $config(cmd,dircp)  \
              [FixFileName $frompwd/[lindex $elem 1]]\
              [FixFileName $destdir]" "" 1  $glob(async)
	  }
        }
      }
      default CantDoThat
    }
  }
}

proc SoftLink { src dst } {
  global config
  if {$config(create_relative_links)} {
    set srcdir [file dirname $src]
    set dstdir [file dirname $dst]
    set srcfile [file tail $src]
    set dstfile [file tail $dst]
    set dstlist [split $dstdir /]
    set srclist [split $srcdir /]
    set dstlen [llength $dstlist]
    set srclen [llength $srclist]
    # Count how many directories are the same in the source and destination paths
    set index 0
    while {([lindex $srclist $index] == [lindex $dstlist $index]) && ($index < $srclen) && ($index < $dstlen)} {
      incr index
    }
    # Build relative link
    set link {}
    for {set dstindex $index} {$dstindex < $dstlen} {incr dstindex} {
      append link ../
    }
    for {set srcindex $index} {$srcindex < $srclen} {incr srcindex} {
      append link [lindex $srclist $srcindex]/
    }
    set from $link$srcfile
    cd $dstdir
    Log [_ "%s : file -s %s" \
	     "$src $dst $srcdir $srcfile $dstdir $dstfile" "$dst $from"]
    if {[Try {file link -symbolic $dst $from} "" 0] == 1 } {
       Try {file link  $dst $from} "" 1
    }
   
  } else {
    Try {file link -symbolic $dst $src} "" 1
  }
}

proc CmdSoftLink {} {
  global glob
  if { $glob(left,pwd) == $glob(right,pwd) } {
    tk_dialog .apop  [_ "Oops That doesn't work"] \
       [_ "The source and destination directories must differ!"] "" 0 [_ "OK"] 
    return {}
  }    
  if {[IsFTP $glob(left,pwd)] || [IsFTP $glob(right,pwd)]} {
    # Well, with FTP we actually can, but we need them both to be
    # on the same remote 
    set l [regexp {([a-z]?)ftp://([^/]*)(.*)} $glob(left,pwd) \
	       match lsftp lftpI ldirectory]
    set r [regexp {([a-z]?)ftp://([^/]*)(.*)} $glob(right,pwd) \
	       match rsftp rftpI rdirectory]
    if {$l && $r && $rftpI == $lftpI} {

    } else {
      CantDoThat
      return
    }
  }
  CmdSoftLink_ $glob(listbox,left).file glob(left,filelist) $glob(left,pwd) $glob(right,pwd)
  CmdSoftLink_ $glob(listbox,right).file glob(right,filelist) $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
}

proc CmdSoftLink_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  global config
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "SoftLink"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      n   -
      d   -
      l   -
      ld  { 
        Log [_ "Softlinking %s to %s" "$frompwd/[lindex $elem 1]" $topwd]
        SoftLink $frompwd/[lindex $elem 1] $topwd/[lindex $elem 1]
      }
      fld -
      fd -
      l -
      ld {
       Log [_ "Softlinking %s to %s" "$frompwd/[lindex $elem 1]" $topwd]
	regexp {ftp://([^/]*)(.*)} $frompwd match ftpI directory
	regexp {ftp://([^/]*)(.*)} $topwd match ftpI tdirectory
	Try {FTP_LinkFile $ftpI \
		 "$directory/[lindex $elem 1]" \
		 "$tdirectory/[lindex $elem 1]"} "" 1
      }
      default CantDoThat
    }
  }
}

proc CmdSoftLinkAs {} {
  global glob
  CmdSoftLinkAs_ $glob(listbox,left).file glob(left,filelist) $glob(left,pwd) $glob(right,pwd)
  CmdSoftLinkAs_ $glob(listbox,right).file glob(right,filelist) $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
}

proc CmdSoftLinkAs_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  global config glob
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "SoftLinkAs"]]} return
    set elem [lindex $filelist $sel]
    set destfile [EntryDialog  "" [_ "Soft-Link As..."] \
		      [_ "Please enter new name for destination link"] \
		      $topwd/[lindex $elem 1] question]
    if {[IsFTP $frompwd] || [IsFTP $destfile]} {
      # Well, with SFTP we actually can, but we need them both to be
      # on the same remote & SFTP (but I said that)
      set l [regexp {([a-z]?)ftp://([^/]*)(.*)} $frompwd \
		 match lsftp lftpI ldirectory]
      set r [regexp {([a-z]?)ftp://([^/]*)(.*)} $destfile \
		 match rsftp rftpI rdirectory]
 #     puts "$l && $r && $rftpI == $lftpI && $lsftp $ldirectory $rdirectory"
      if {$l && $r && ($rftpI == $lftpI) } {
      } else {
	CantDoThat
	return
      }
    }
    if {$destfile != "" } {
      Log [_ "Softlinking %s to %s" "$frompwd/[lindex $elem 1]" $destfile]
      switch [lindex $elem 2] {
	n   -
	d   -
	l   -
	ld  {
          SoftLink $frompwd/[lindex $elem 1] $destfile
        }
	fld -
	fd -
	fn -
	ld {
	  Log [_ "Softlinking %s to %s" "$frompwd/[lindex $elem 1]" $destfile]
	  Try {FTP_LinkFile $rftpI \
		   "$ldirectory/[lindex $elem 1]" \
		   "$rdirectory"} "" 1
	}
	default {CantDoThat}
      }
    }
  }
}


proc CmdMove {} {
  global glob
  if { $glob(left,pwd) == $glob(right,pwd) } {
    tk_dialog .apop [_ "Oops That doesn't work"] \
       [_ "Please set different source and destination directories!"] "" 0 [_ "OK"] 
    return {}
  }    
#  if {[IsFTP $glob(left,pwd)] || [IsFTP $glob(right,pwd)]} {
#    FTP_Move
#    return
#  }
  CmdMove_ $glob(listbox,left).file glob(left,filelist) $glob(left,pwd) $glob(right,pwd)
  CmdMove_ $glob(listbox,right).file glob(right,filelist) $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
}

proc CmdMove_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  global config glob
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Move"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      d   -
      l   -
      ld  -
      n   { 
        if {[CheckWhoOwns $frompwd/[lindex $elem 1] move] && \
		[CheckWhoOwns $topwd/[lindex $elem 1] overwrite]} {
          Log [_ "Moving %s to %s" $frompwd/[lindex $elem 1] $topwd]
          Try { file rename -force -- \
		    $frompwd/[lindex $elem 1] $topwd } "" 1 $glob(async)
        }
      }
      fn  -
      fld -
      fl  -
      fd {
	set rf [regexp {ftp://([^/]*)(.*)} \
		    $frompwd/[lindex $elem 1] match ftpIf ffile]
	set rt [regexp {ftp://([^/]*)(.*)} \
		    $topwd/[lindex $elem 1] match ftpIt tfile]
	if {$rf == 1 && $rt == 1 && $ftpIf == $ftpIt } {
	  Try { FTP_Rename $ftpIf $ffile $tfile } "" 1
	  set glob(forceupdate) 1	  
	} else {
	  # to or from either different ftp locations or ftp & local file
	  # do a copy followed by a delete
	  # maybe later...
          CantDoThat
          return
	}  
      }
	
      default CantDoThat
    }
  }
}

proc CmdDelete {} {
  global glob
  CmdDelete_ $glob(listbox,left).file glob(left,filelist) $glob(left,pwd) $glob(right,pwd) left
  CmdDelete_ $glob(listbox,right).file glob(right,filelist) $glob(right,pwd) $glob(left,pwd) right
  UpdateWindow both
  set glob(forceupdate) 0
}

proc CmdDelete_ { listb_name filelist_var frompwd topwd inst } {
  # For speed, (hopefully) we'll call by reference...
  global config glob
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Delete"]]} return
    set elem [lindex $filelist $sel]
    set type [lindex $elem 2]
    set ford [switch $type {
      d -     
      fd {if {$config(ask,dir_delete)} {expr {{directory tree}}}} 
      default {if {$config(ask,file_delete)} {expr {{file}}}}
    }]
    switch $type {
      l   -
      ld  -
      n   -
      d { 
        set ask 0
        if {$ford != ""} {
          set ask [tk_dialog_fr .apop [_ "Sure?"] \
		       [_ "OK to delete %s ?" "$ford $frompwd/[lindex $elem 1]"]\
		       "" 1 [_ "Yes"] [_ "No"]]
        }
        if {$ask == 0} {
          if {[CheckWhoOwns $frompwd/[lindex $elem 1] delete]} {
            Log [_ "Deleting %s" $frompwd/[lindex $elem 1]]
            Try { file delete -force -- $frompwd/[lindex $elem 1] } "" 1
          }
        }
      }
      fn  -
      fld -
      fl  {
        set ask 0
        if {$ford != "" } {
          set ask [tk_dialog_fr .apop [_ "Sure?"] \
		       [_ "OK to delete %s ?" \
			    "$ford $frompwd/[lindex $elem 1]"] "" 1 \
		       [_ "Yes"] [_ "No"]]
        }
        if {$ask == 0} {
          Log [_ "Deleting %s" $frompwd/[lindex $elem 1]]
          set r [regexp {ftp://([^/]*)(.*)} \
		     $frompwd/[lindex $elem 1] match ftpI file]
          if {$r} {
            set glob(forceupdate) 1
            Try { FTP_Delete $ftpI "$file" } "" 1
          }
        }
      }
      fd  {
        set ask 0
        if {$ford != "" } {
          set ask [tk_dialog_fr .apop \
		       [_ "Sure?"] \
		       [_ "OK to delete %s ?" "$ford $frompwd/[lindex $elem 1]"]\
		       "" 1 [_ "Yes"] [_ "No"]]
        }
        if {$ask == 0} {
          Log [_ "Deleting %s" $frompwd/[lindex $elem 1]]
          set glob(forceupdate) 1
          Try { DeleteFTPRecursive "$frompwd/[lindex $elem 1]" $inst } "" 1
        }
      }
      default CantDoThat
    }
  }
}

proc CmdView {} {
#  puts "CmdView"
  global glob
  CmdView_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd) left
  CmdView_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) right
}

proc CmdView_ { listb_name filelist_var frompwd topwd inst } {
  # For speed, (hopefully) we'll call by reference...
#  puts "CmdView_"
  global glob config env
  upvar $filelist_var filelist
  set filenamelist {}
  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "View"]]} return
    set fileelem [lindex $filelist $sel]
    switch [lindex $fileelem 2] {
      l   -
      n   { lappend filenamelist $frompwd/[lindex $fileelem 1] }
      fd  -
      fld -
      ld  - 
      d   { NewPwd $inst $glob($inst,pwd)/[lindex $fileelem 1]
	         UpdateWindow $inst
	         return
          }
      fn  -
      fl  {
	         set r [regexp {ftp://([^/]*)(.*)} \
			    $glob($inst,pwd) match ftpI directory]
	         if {$r == 0} { 
	           PopError [_ "Can't parse %s as ftp URL" $glob($inst,pwd)] 
         	} else { 
	            set r 0
	            if { ! [file exists $glob(tmpdir)] } {
	              set r [Try { file mkdir $glob(tmpdir) } "" 1]
	            }
	            if { !$r } {
	              set size [lindex $fileelem 3]
	              if {[lindex $fileelem 2] == "fl"} {set size -1}
	              set r [Try { 
	                FTP_GetFile $ftpI "$directory/[lindex $fileelem 1]" \
			    "$glob(tmpdir)/[lindex $fileelem 1]"\
			    $size 0 } "" 1]
	                if {$r == 0} {
#	                  puts "$glob(tmpdir)/[lindex $fileelem 1]"
	                  ViewAny [regsub -all {\ }\
				      "$glob(tmpdir)/[lindex $fileelem 1]" {\\ }]
	                  set glob(havedoneftp) 1 
			        }
	            }
         	}
          }
          default CantDoThat
    }
  }
  if {$filenamelist != {}} {
    ViewAny $filenamelist
  }
}

proc CmdViewAsText {} {
  global glob
  CmdViewAsText_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd)
  CmdViewAsText_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd)
}

proc CmdViewAsText_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  upvar $filelist_var filelist
  set filenamelist {}
  foreach sel [$listb_name curselection] {
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      n   { ViewText $frompwd/[lindex $elem 1] }
      ld  -
      d   { PopInfo [_ "Can't view directory %s in the text viewer" \
			 "$frompwd/[lindex $elem 1]"] }
      default CantDoThat
    }
  }
}

proc CmdCheckSize {} {
  global glob
  CmdCheckSize_ $glob(listbox,left).file glob(left,filelist)\
      $glob(left,pwd) $glob(right,pwd) left
  CmdCheckSize_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) right
}

proc CmdCheckSize_ { listb_name filelist_var frompwd topwd inst } {
  # For speed, (hopefully) we'll call by reference...
  global glob config env
  upvar $filelist_var filelist
  set filenamelist {}
  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "View"]]} return
    set fileelem [lindex $filelist $sel]
    switch [lindex $fileelem 2] {
      d   -
      ld  -
      l   -
      n   { lappend filenamelist [Esc $frompwd/[lindex $fileelem 1]] }
      default CantDoThat
    }
  }
  if {$filenamelist != {}} {
    set r [catch {eval eval exec $config(cmd,du) $filenamelist} out]
    if {$r} {
      PopError $out
    } else {
      ViewString [_ "Output from %s" "du"] out ""
    }
  }
}

proc CmdWhatIs {} {
  global glob
  CmdWhatIs_ $glob(listbox,left).file glob(left,filelist)\
      $glob(left,pwd) $glob(right,pwd)
  CmdWhatIs_ $glob(listbox,right).file glob(right,filelist)\
      $glob(right,pwd) $glob(left,pwd)
}

proc CmdWhatIs_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  upvar $filelist_var filelist
  set filenamelist {}
  foreach sel [$listb_name curselection] {
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      n   -
      ld  -
      d   { Try { PopInfo [exec file "$frompwd/[lindex $elem 1]"] } "" 1 
          }
      default CantDoThat
    }
  }
}

proc CmdEdit {} {
  global glob
  CmdEdit_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd)
  CmdEdit_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd)
}

proc CmdEdit_ { listb_name filelist_var frompwd topwd } {
  global env config
  # For speed, (hopefully) we'll call by reference...
  upvar $filelist_var filelist
  set filenamelist {}

  foreach sel [$listb_name curselection] {
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      n   { lappend filenamelist [FixFileName  $frompwd/[lindex $elem 1]] }
      ld  -
      d   { PopInfo [_ "Can't edit directory %s in the text editor" \
			 "$frompwd/[lindex $elem 1]"] }
      default CantDoThat
    }
  }
  if {$filenamelist != {}} {
    # Three eval's in a row... New record? Yuck..
    Try { eval eval eval exec [format $config(editor)  $filenamelist] & } \
	 [_ "Can't start editor"] 1
  }
}

proc CmdQEdit {} {
  global glob
  CmdQEdit_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd)
  CmdQEdit_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd)
}

proc CmdQEdit_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  global glob
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Q-Edit"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      n   { 
            set r [Try {EditText "$frompwd/[lindex $elem 1]" ""} \
		       [_ "Error editing %s" $frompwd/[lindex $elem 1]] 1]
            if {$r != 0} { catch { destroy .toplevel_$glob(toplevelidx) } }
          }
      ld  -
      d   { PopInfo [_ "Can't edit directory %s in the text editor" \
			 "$frompwd/[lindex $elem 1]"] }
      default CantDoThat
    }
  }
}


proc CmdRename {} {
  global glob
  CmdRename_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd)
  CmdRename_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
  set glob(forceupdate) 0
}

proc CmdRename_ { listb_name filelist_var frompwd topwd } {
  # For speed, (hopefully) we'll call by reference...
  global config glob
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Rename"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      d   -
      l   -
      ld  -
      n   { 
        if {[CheckWhoOwns $frompwd/[lindex $elem 1] rename]} {
          set newname [EntryDialog "" \
			   [_ "Rename"] \
			   [_ "Please enter new name."] \
			   $frompwd/[lindex $elem 1] question]
          if {$newname != ""} {
            if {[CheckWhoOwns $newname overwrite]} {
              Log [_ "Renaming %s to %s" $frompwd/[lindex $elem 1] $newname]
              Try { file rename -force -- $frompwd/[lindex $elem 1] $newname }\
		  "" 1
            }
          }
        }
      }
      fl  -
      fld -
      fd  -
      fn  {
        set newname [EntryDialog  "" [_ "Rename"] \
			 [_ "Please enter new name."] \
			 $frompwd/[lindex $elem 1] question]
        if {$newname != ""} {
          Log [_ "Renaming %s to %s" "$frompwd/[lindex $elem 1]" $newname]
          set r [regexp {ftp://([^/]*)(.*)} $frompwd/[lindex $elem 1]\
		     match ftpI oldftpname]
          if {!$r} {
            PopError [_ "Error in URL %s" $frompwd/[lindex $elem 1]]
          } else {
            set r [regexp {ftp://([^/]*)(.*)} $newname match ftpI newftpname]
            if {!$r} {
              PopError [_ "Error in URL %s" $newname]
            } else {
              Try { FTP_Rename $ftpI $oldftpname $newftpname } "" 1
              set glob(forceupdate) 1
            }
          }
        }
      }
      default CantDoThat
    }
  }
}


proc CmdUnArc {} {
  global glob
  CmdUnArc_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd)
  CmdUnArc_ $glob(listbox,right).file glob(right,filelist)\
      $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
}

proc CmdUnArc_ { listb_name filelist_var frompwd topwd } {
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "UnArchive"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      ld  -
      d   { PopError [_ "Can't unarchive directory %s..." [lindex $elem 1]] }
      l   -
      n   { Log [_ "Unarchiving %s to %s" $frompwd/[lindex $elem 1] $topwd]
            UnArcAny $frompwd/[lindex $elem 1] $topwd 
          }
      default CantDoThat
    }
  }
}

proc CmdUnPack {} {
  global glob
  CmdUnPack_ $glob(listbox,left).file glob(left,filelist)\
      $glob(left,pwd) $glob(right,pwd)
  CmdUnPack_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
}

proc CmdUnPack_ { listb_name filelist_var frompwd topwd } {
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "UnPack"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      ld  -
      d   { PopError [_ "Can't UnPackhive directory %s..." [lindex $elem 1]] }
      l   -
      n   { Log [_ "Unpacking %s" $frompwd/[lindex $elem 1]]
            UnPackAny $frompwd/[lindex $elem 1] $topwd }
      default CantDoThat
    }
  }
}


proc CmdArc {} {
  global glob
  CmdArc_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd)
  CmdArc_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd)
  UpdateWindow both
}

proc CmdArc_ { listb_name filelist_var frompwd topwd } {
  global config glob
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Archive"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      n   { 
        Log [_ "Packing %s" $frompwd/[lindex $elem 1]]
#        Try { eval exec [format $config(cmd,pack) \
#	    {$frompwd/[lindex $elem 1]}] } "" 1  $glob(async)
        Try { eval eval exec [format $config(cmd,pack) \
				  [Esc $frompwd/[lindex $elem 1]]] }\
	    "" 1  $glob(async)
      }
      ld  -
      d   { 
        Log [_ "Archiving %s" $frompwd/[lindex $elem 1]]
        if {$config(cmd,archive) == "tar+gz {%s}"} {
          Try { cd $frompwd; exec tar cf - [lindex $elem 1] | \
		    gzip > [lindex $elem 1].tar.gz } "" 1 $glob(async)
        } else {
          Try { cd $frompwd; eval eval exec \
		    [format $config(cmd,archive) \
			 [Esc [lindex $elem 1]] \
			 $topwd  [Esc [lindex $elem 1]] ] }\
	      "" 1 $glob(async)
        }
      }
      default CantDoThat
    }
  }
}

proc CmdPrint {} {
  global glob
  CmdPrint_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd)
  CmdPrint_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd)
}

proc CmdPrint_ { listb_name filelist_var frompwd topwd } {
  global config
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Print"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      n   { Log [_ "Printing %s" $frompwd/[lindex $elem 1]]
            Try { eval eval exec \
		      [format $config(cmd,print) \
			   [Esc $frompwd/[lindex $elem 1]]] } "" 1 }
      ld  -
      d   { PopError [_ "Can't print directories!!"] }
      default CantDoThat
    }
  }
}

proc CmdMakeDir {} {
  global config glob
  if { [info exists glob(whichdir)] } {
    set newdir [$glob(win,$glob(whichdir)).entry_dir get]
  } else {
    set newdir [EntryDialog "" [_ "Directory name?"] \
		    [_ "Please enter the name of the new directory.\
                        Another way of creating directories is to enter\
                        the name the new directory in one\
                        of the directory entries and then pressing the\
                        MkDir button"] "$glob(left,pwd)" question]
    if {$newdir == ""} return
  }
  Log [_ "Creating directory %s" $newdir]
  set r [regexp {ftp://([^/]*)(.*)} $newdir match ftpI dir]
  if {$r} {
    Try { FTP_MkDir $ftpI "$dir" } "" 1
    set glob(forceupdate) 1
  } else {
    Try { file mkdir $newdir } "" 1
  }
  UpdateWindow both
  set glob(forceupdate) 0
}

proc CmdSelect {} {
  global glob
  if { [info exists glob(whichdir)] } {
    set pat [$glob(win,$glob(whichdir)).entry_dir get]
    set inst $glob(whichdir)
  } else {
    PopInfo [_ "Please enter a selection pattern in one of the\
                directory entries and then press the Select button"]
    return
  }
  $glob(win,$inst).entry_dir delete 0 end
  $glob(win,$inst).entry_dir insert end $glob(${inst},pwd)
  set pat [file tail $pat]
  set i 0
  foreach elem $glob($inst,filelist) {
    if {[string match $pat [lindex $elem 1]]} {
      $glob(listbox,$inst).file selection set $i
    }
    incr i
  }
  propagateSelection $glob(listbox,$inst).file
  UpdateStat
}

proc CmdCSelect {} {
  global glob
  set cmd [EntryDialog  "" [_ "Contents-select"] \
	       [_ "Make sure you have selected the files you want to\
                   search in, then please edit this command to do a\
                   contents-select."] "grep -i "]
  if { $cmd == "" } return
  CmdCSelect_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd) $cmd
  CmdCSelect_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) $cmd
  UpdateStat
}

proc CmdCSelect_ { listb_name filelist_var frompwd topwd cmd } {
  upvar $filelist_var filelist

  foreach sel [$listb_name curselection] {
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      n   { 
            set r [catch { eval exec $cmd {$frompwd/[lindex $elem 1]} } outp]
            if { $r != 0 } {
              $listb_name selection clear $sel
            }
          }
      default CantDoThat
    }
  }
}

proc CmdRunCmd {} {
  global glob
  CmdRunCmd_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd) 
  CmdRunCmd_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) 
}

proc CmdRunCmd_ { listb_name filelist_var frompwd topwd } {
  global glob
  upvar $filelist_var filelist

  set fl {}
  foreach sel [$listb_name curselection] {
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      d   -
      ld  -
      l   -
      n   { lappend fl [lindex $elem 1] }
      default CantDoThat
    }
  }
  if { $fl == "" } return
  set cmd \
      [EntryDialog  "" [_ "Run command"] \
	   [_ "Edit the command line. Ctrl-A\
               takes you to the start of the line."] \
	   " $fl"]
  if { $cmd == "" } return
  LogSilent [_ "Running command:\n%s" $cmd]
  if {$glob(async)} {
    catch { cd $frompwd; eval exec $cmd & } out
    return
  } else {
    catch { cd $frompwd; eval exec $cmd } out
  }
  if { $out != "" } {
    ViewString [_ "Output from command"] out ""
  }
  UpdateWindow both
}

proc CmdForEach {} {
  global glob
  CmdForEach_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd) 
  CmdForEach_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) 
}

proc CmdForEach_ { listb_name filelist_var frompwd topwd } {
  global glob config
  upvar $filelist_var filelist

  set fl {}
  foreach sel [$listb_name curselection] {
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      d   -
      ld  -
      l   -
      n   { lappend fl [lindex $elem 1] }
      default CantDoThat
    }
  }
  if { $fl == "" } return
  if {$glob(async)} {
    PopError [_ "This command does not support asynchronous execution"]
    return
  }
  if {![info exists glob(foreach,cmd)] || $glob(foreach,cmd) == {}} {
    set glob(foreach,cmd) {echo '%s'}
  }
  set glob(foreach,cmd) \
      [EntryDialog "" [_ "Foreach"] \
	   [_ "Enter command to run on each of the selected files.\
               The file will show up in the '%s'. You can use pipes\
               and other bourne-shell syntax elements since the\
               commands will each run in a separate bourne shell." "%s"]\
	   $glob(foreach,cmd)]
  if { $glob(foreach,cmd) == "" } return

  set output {}
  foreach k $fl {
    if {[CheckAbort [_ "ForEach"]]} return
    set cmd [format "$glob(foreach,cmd)" $k]
    append output "\$ $cmd\n"
    LogStatusOnly [_ "Running %s ..." $cmd]
    catch { cd $frompwd ; uplevel 0 exec $config(cmd,sh)  "$cmd" } out
    append output "$out\n"
  }
  ViewString [_ "Output from commands"] output ""
  UpdateWindow both
}

proc CmdRecurseCommand { inst } {
  global glob config

  set dir $glob($inst,pwd)

  if { [IsFTP $dir] } {
    CantDoThat
    return
  }

  if {$glob(async)} {
    PopError [_ "This command does not support asynchronous execution"]
    return
  }
  if {![info exists glob(foreach,cmd)] || $glob(foreach,cmd) == {}} {
    set glob(foreach,cmd) {echo '%s'}
  }
  set e [EntryDialogDouble [_ "Run Command Recursively"] \
	     [_ "Enter command to run on each file.\
                 The file will show up in the '%s'.\
                 You can use pipes and other bourne-shell syntax\
                 elements since the commands will each run in a\
                 separate bourne shell." "%s" ]\
	     [_ "Enter pattern to match for when recursing into directories."]\
	     [_ "This command will run a '%s' of the type\n\n    %s\n\nto\
                 recurse into the current directory (%s).\
                 The command will then be run on all files\
                 from this find.\n\nSee also: manpage for \
                 the find command.\n\nUse tab to go to next entry.\
                 Return or the OK button starts execution."\
                 "find" "$ find <directory> -type f -name <pattern> -print"\
                 $dir] $glob(foreach,cmd) {*}]
  if { $e == "" } return

  set glob(foreach,cmd) [lindex $e 0]
  set pattern [lindex $e 1]
  set output {}

  # expand file list with find outputs
  set r [catch {cd $dir ; uplevel 0 \
		    exec $config(cmd,find) . -type f -name $pattern -print} out]
  if {$r} {
    PopError "$out"
    return
  }
  set fl [split $out \n]
  foreach k $fl {
    if {[CheckAbort [_ "Recurse Command"]]} return
    set cmd [format "$glob(foreach,cmd)" $k]
    append output "\$ $cmd\n"
    LogStatusOnly [_ "Running %s ..." $cmd]
    catch { cd $dir ; uplevel 0 exec $config(cmd,sh)  "$cmd" } out
    append output "$out\n"
  }
  ViewString [_ "Output from commands"] output ""
  UpdateWindow both
}


proc CmdDiff {} {
  global glob
  CmdDiff_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd) left
  CmdDiff_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) right
}

proc CmdDiff_ { listb_name filelist_var frompwd topwd inst } {
    global glob
    upvar $filelist_var filelist
    global config

    set null 1
    set file1 ""
    set file2 ""
    foreach sel [$listb_name curselection] {
	set null 0
	set elem [lindex $filelist $sel]
	switch [lindex $elem 2] {
	    ld  -
	    d   -
	    l   -
	    n   { 
		if {$file1 == ""} { 
		    set file1 $frompwd/[lindex $elem 1] 
		} else {
		    if { $file2 != "" } { 
			PopError [_ "Please select one or two\
                                    files or directories for diffing."]
			return
		    }
		    set file2 $frompwd/[lindex $elem 1] 
		    # But which was first?
		    if {$inst == $glob(select_pry_lr) && 
			$sel == $glob(select_pry_s)} { } else {
			set tmp $file1
			set file1 $file2
			set file2 $tmp
		    }
		}
	    }
	    default CantDoThat
	}
    }

    if {$null} return

    if {$file2 == "" && $glob(select_pry_lr) != "nil"} {
	set sel [split $glob(select_pry_s)]
	if {[llength $sel] == 1} {
	    set elem2 [lindex $glob($glob(select_pry_lr),filelist) $sel]
	    set file2  $glob($glob(select_pry_lr),pwd)/[lindex $elem2 1]
	} else {
	    PopError [_ "Please select one or two\
                         files or directories for diffing."]
	    return
	}
    }     
    if {! ($file2 == "")} {
	set tmp $file1
	set file1 $file2
	set file2 $tmp
    }
   LogStatusOnly "$file1 $file2" 
    set r [catch { eval eval \
		      exec [format $config(cmd,diff) [Esc $file1] [Esc $file2]] \
		     } outp]
    if { $r != 0 } {
	ViewString [_ "Diffing %s and %s" $file1 $file2] outp "diff.log"
    } else {
	PopInfo [_ "No difference between\n\n%s\n\nand\n\n%s" $file1 $file2]
    }
}

proc CmdCreateEmptyFile { inst } {
  global glob config
  set start_entry $glob($inst,pwd)
  set newfile [EntryDialog  "" [_ "Create New File"] \
		   [_ "Please enter the name of the new file."] \
		   $start_entry question]

  if {$newfile != ""} {
    Log [_ "Creating new file %s" $newfile]
    Try { exec $config(cmd,touch) "$newfile" } "" 1
    UpdateWindow both
  }
}

proc DeleteFTPRecursive { dir inst } {
  global glob config
  set r [regexp {ftp://([^/]*)(.*)} $dir match ftpI directory]
  Log [_ "FTP recursive delete: Entering %s" $dir]
  NewPwd $inst $dir
  UpdateWindow $inst
  foreach elem $glob($inst,filelist) {
    switch [lindex $elem 2] {
      fn  -
      fld -
      fl  {
        Log [_ "Deleting %s" $dir/[lindex $elem 1]]
        FTP_Delete $ftpI "$directory/[lindex $elem 1]"
      }
	fd { 
	    
	  if { [lindex $elem 1] != "." } {
	      if { [lindex $elem 1] != ".." } {
		  DeleteFTPRecursive "$dir/[lindex $elem 1]" $inst
	      }
	  }
      }
      default {
        error [_ "Unexpected file type in FTP recursive delete"]
      }
    }
  }
  Log [_ "Deleting %s" $dir]
  NewPwd $inst "$dir/.."
  UpdateWindow $inst
  FTP_RmDir $ftpI "$directory"
}

proc CopyFromFTPRecursive { fromdir todir inst resume } {
  global glob config
  if {[CheckAbort [_ "CopyFromFTPRecursive"]]} return
  set r [regexp {ftp://([^/]*)(.*)} $fromdir match ftpI directory]
  set dir [file tail $directory]
  Log [_ "FTP recursive copy: Creating %s" $todir/$dir]
  file mkdir "$todir/$dir"
  Log [_ "FTP recursive copy: Entering %s" $fromdir]
  NewPwd $inst $fromdir
  UpdateWindow $inst
  foreach elem $glob($inst,filelist) {
    if {[CheckAbort [_ "CopyFromFTPRecursive"]]} return
    switch [lindex $elem 2] {
      fld -
      fl {
        Log [_ "Skipping %s - link" $fromdir/[lindex $elem 1]]
      }
      fd {
	  if { [lindex $elem 1] != "." } {
	      if { [lindex $elem 1] != ".." } {
		  CopyFromFTPRecursive "$fromdir/[lindex $elem 1]" \
		      "$todir/$dir" $inst $resume
	      }
	  }
      }
      fn {
        Log [_ "FTP recursive copy: Copying %s -> %s bytes)" \
             $fromdir/[lindex $elem 1] \
             "$todir/$dir/[lindex $elem 1] ([lindex $elem 3]"]
        FTP_GetFile $ftpI "$directory/[lindex $elem 1]" \
	    "$todir/$dir/[lindex $elem 1]" "[lindex $elem 3]" $resume
      }
      default {
        error [_ "Unexpected file type in FTP recursive copy"]
      }
    }
  }
}

proc CopyToFTPRecursive { fromdir todir inst } {
  global glob config
  if {[CheckAbort [_ "CopyToFTPRecursive"]]} return
  set r [regexp {ftp://([^/]*)(.*)} $todir match ftpI directory]
  set dir [file tail $fromdir]
  Log [_ "FTP recursive copy: Creating %s" $todir/$dir]
  FTP_MkDir $ftpI "$directory/$dir"
  Log [_ "FTP recursive copy: Entering %s" $fromdir]
  NewPwd $inst $fromdir
  UpdateWindow $inst
  foreach elem $glob($inst,filelist) {
    if {[CheckAbort [_ "CopyToFTPRecursive"]]} return
    switch [lindex $elem 2] {
      ld -
      l {
        Log [_ "Skipping %s - link" $fromdir/[lindex $elem 1]]
      }
      d {
 	  if { [lindex $elem 1] != "." } {
	      if { [lindex $elem 1] != ".." } {
		  CopyToFTPRecursive "$fromdir/[lindex $elem 1]" \
		      "$todir/$dir" $inst
	      }
	  }
      }
      n {
        Log [_ "FTP recursive copy: Copying %s -> %s bytes)"\
		 $fromdir/[lindex $elem 1] \
		 "$todir/$dir/[lindex $elem 1] ([lindex $elem 3]"]
        FTP_PutFile $ftpI "$fromdir/[lindex $elem 1]" \
	    "$directory/$dir/[lindex $elem 1]" "[lindex $elem 3]"
      }
      default {
        error [_ "Unexpected file type in FTP recursive copy"]
      }
    }
  }
}

proc CmdFind { inst } {
  global glob config
  set findname [EntryDialog  "" [_ "Find..."] \
		    [_ "Please enter substring of filename\
                        to search for in %s and below." $glob($inst,pwd)] \
		    "" "question"]
  if {$findname == ""} return
  if {[IsFTP $glob($inst,pwd)]} {
    regexp {ftp://([^/]*)(.*)} $glob($inst,pwd) match ftpI directory
    set r [catch {FTP_CD $ftpI $directory} out]
    if {$r != 0} {
      PopError $out
      return
    }
    LogStatusOnly [_ "Searching, please wait..."]
    set r [catch {FTP_DoSearch $ftpI $findname} out]
    LogStatusOnly [_ "Searching, please wait...done"]
    if {$r} {
      PopError [_ "FTP search error: %s" $out]
      return
    }
    ViewString [_ "FTP search results"] out ""
    return
  }
  set r [catch {cd $glob($inst,pwd)} out]
  if {$r} {
    PopError "$out"
    return
  }
  set pat "\\\"*$findname*\\\""
  LogStatusOnly [_ "Searching, please wait..."]
  set r [catch {eval  \
		    exec $config(cmd,sh)  \
		    \"find . -name $pat -print 2> /dev/null\"} out]
  eval {puts "exec $config(cmd,sh)  \
		    find . -name $pat -print 2> /dev/null"}
  LogStatusOnly [_ "Searching, please wait...done"]
#  if {$r} {
#    PopError "Search error: $out"
#    return
#  }
  set out [split $out "\n"]
  if {[string range [lindex $out end] 0 4] == "child"} {
    set out [lrange $out 0 [expr [llength $out] - 2]]
  }
  if {[string trim $out] == ""} {
    PopInfo [_ "No file found"]
    return
  }
  FindDialog $out $inst
}


proc CmdChmod {} {
  global glob
  CmdChmod_ $glob(listbox,left).file glob(left,filelist) \
      $glob(left,pwd) $glob(right,pwd) left
  CmdChmod_ $glob(listbox,right).file glob(right,filelist) \
      $glob(right,pwd) $glob(left,pwd) right
}

proc CmdChmod_ { listb_name filelist_var frompwd topwd inst } {
  # For speed, (hopefully) we'll call by reference...
  global config glob
  upvar $filelist_var filelist
  set fl {}

  foreach sel [$listb_name curselection] {
    if {[CheckAbort [_ "Chmod"]]} return
    set elem [lindex $filelist $sel]
    switch [lindex $elem 2] {
      l   -
      ld  -
      n   -
      d   {
	incr ftpFlag 0
	if {$fl == {}} {
	  set mode [file attributes $frompwd/[lindex $elem 1] -permissions]
	}
      }
      default {
	# Must be an nFTP, maybe we can do it... up to the handler and remote
	incr ftpFlag
	if {$fl == {}} {
	  set mode [GetModeFromString [lindex $elem 5]]
#	  puts "got $mode from >[lindex $elem 5]<"
	}
      }
    }
    lappend fl [Esc $frompwd/[lindex $elem 1]]
  }
  if {$fl != ""} {
    set arg [ChmodDialog "$frompwd/[lindex $elem 1]" $mode]
    if {$arg != ""} {
      if { $ftpFlag } {
	# for ftp we will use a seperate call for each file...
	foreach file $fl {
	  set r [regexp {ftp://([^/]*)(.*)} $file match ftpI fl]
	  if {$r == 0} { 
	    PopError [_ "Can't parse %s as ftp URL" $frompwd] 
	  } else {
	    set r [FTP_Chmod $ftpI $arg $fl]
	    if { $r != "0" && $r != "" } {
	      PopError $r
	      break
	    }
	  }
	}
      } else {
	Try { eval eval exec $config(cmd,chmod) $arg $fl } "" 1 $glob(async)
      }
      ForceUpdate $inst
    }
  }
}

proc CmdGetHttp { inst } {
  global glob config
  
  if {[IsFTP $glob($inst,pwd)]} {
    PopInfo [_ "You can only download HTTP files to a non-FTP directory"]
    return
  }

  if {![info exists glob(http,lasturl)]} {
    set glob(http,lasturl) {}
  }
  set URL $glob(http,lasturl)

  while { 1 } {
    set URL [EntryDialog  "" [_ "Get HTTP File"] \
		 [_ "Please enter HTTP URL to download"] $URL]
    if {$URL == ""} {
      return
    }
    
    if {[string range $URL 0 6] != "http://" } {
      set URL "http://$URL"
    }
    
    set r [regexp {http://([^/]*)(.*)} $URL match host_and_port location]
    if {!$r} {
      PopError [_ "Could not parse %s as an HTTP URL" $URL]
      continue
    } 
    if {$location == ""} {
      append URL /
    }
    set glob(http,lasturl) $URL
    break
  }

  set filename [file tail $URL]
  if {[string range $URL end end] == "/"} {
    append filename .html
  }
  set filename [EntryDialog  "" [_ "Get HTTP File"] \
		    [_ "Please edit filename to save to.\n(URL: %s)" $URL] \
		    $filename]
  if {$filename == ""} {
    return
  }
  cd $glob($inst,pwd)
  HTTP_Get $URL $filename
  UpdateWindow both
}

