#############################################################################
# Display size and placement code.
#############################################################################

# These routines probe the display and return a list of the displays and
# their locations on the current screen.  To use, just call:
# getDisplays
# The return is a list of lists.  Each entry is a display and consists of
# {Width Heigth minx miny}  where minx and miny may be negative.  The code 
# should find all displays on the screen (i.e. there is no limit).
# You may notice small windows flashing on the various displays as the
# code probes them.  The information returned defines the visible area; it
# is, at lest sometimes, possible to put windows completle or partiualy off
# screen.  
# This code was done support tooltips where we always want to see
# the 'tip' AND we want it on the same display as the mouse.
#
# As an aside, it would have been nice if there were a proc to return this
# info.  Also, it would be nice if there was a common way to fill the screen.
# Also, we note that the 'winfo screen...' procs seem to return information
# for the prymary display (not the screen and not the display the window is on).
# We use the notation 'screen' to mean the whole, possibly multi head
# thing and 'display' to mean a monitor in that 'screen'.

proc findNoMansLand { result } {
  # This routine takes a list describing the display (minX maxX minY maxY)
  # and determins if there are edges that can not be used (like the task bar)
  # It returns an adjusted 'result'
  catch "destroy .ba"
  set w [toplevel .ba]
  wm overrideredirect $w 1
  set sr 256
  set mn 1
  set X x
  set xmin [lindex $result 0]
  set xmax [expr {[lindex $result 1] - $sr + 1}]
  set ymin [lindex $result 2]
  set ymax [expr {[lindex $result 3] - $sr + 1}]
  set xprob [expr {$sr + $xmin}]
  set yprob [expr {$sr + $ymin}]
  foreach max $result \
      direction { 1 -1 1 -1} \
      geo "$sr$X$mn+$xmin+$yprob\
	       $sr$X$mn+$xmax+$yprob \
	       $mn$X$sr+$xprob+$ymin \
	       $mn$X$sr+$xprob+$ymax" \
      xloc "$xmin $xmax $xprob $xprob" \
      yloc "$yprob $yprob $ymin $ymax " {
	wm geometry $w $geo
	set lv 0
	update
#	puts "$geo [winfo geometry $w ] xloc $xloc yloc $yloc max $max"
	while { [winfo containing $max $yloc] != $w && \
		[winfo containing $xloc $max] != $w } {
#	  puts "1$direction $max >[winfo containing $max $yloc]<"
	  incr max $direction
	  if {[incr lv] > $sr } { break}
	}
	lappend out $max
#	puts "$out $max"
      }
  destroy $w
#  puts "$out"
  return $out
}

proc probeDisplay { w x y } {
  global debug
  set xs 100
  set ys 100
  set X x
#  puts "start [winfo geometry $w] "
  wm geometry $w $xs$X$ys+$x+$y
#  puts "try $xs$X$ys+$x+$y"
  if { [string index [winfo server $w] 0] == "X" } {
    wm attributes $w -fullscreen 0
  } else {
#    wm state $w normal
  }
    wm deiconify $w
  # raise $w
  # Well, this is messy.  The window manager runs on its own clock!
  # That means we have to wait for it AND the only clue we have
  # is that the geometry changes.  Also, it seems to require
  # the update thing.  Further, it needs (well at least this works)
  # the fullscreen thingy to kick it off.  For what its worth,
  # the first loop usually is at most twice, the second, up to 8.
  # That is on a fedora 14 X-window system.

  # reset the fail safe counters...
  set loop [set loop2 0]
  while { 1 } {
    set sz "[regsub -all {[x+]} [winfo geometry $w] { }]"
    if { [lindex $sz 0] != $xs || [lindex $sz 1] != $ys } {
      update
      if { [incr loop] > 20} { 
	puts "loop1 $sz > $xs $ys"
	return ""
      }
#      puts "P1 $sz $xs $ys $x $y $loop"
      continue
    }
    break
  }
  update
  set sz "[regsub -all {[x+]} [winfo geometry $w] { }]"
  
  # If the WM grossly ignored our request, fail without deiconifying...
  set tx [expr {$x + $xs / 2}]
  set ty [expr {$y + $ys / 2}]
  if {$tx < [lindex $sz 2]  || $ty < [lindex $sz 3] || \
	  $tx > ([lindex $sz 0] + [lindex $sz 2]) || \
	  $ty > ([lindex $sz 1] + [lindex $sz 3]) } {
    if {$debug } { puts "out $sz vs $x $y $tx $ty"}
    return ""
  }

#  wm deiconify $w
  incr loopx 0
  while { [winfo containing  $tx $ty] == "" } {
    update
    if {[incr loopx] > 20 } {
      if {$debug } {
	puts "!contain $sz > $x $y $tx $ty xs,ys $xs,$ys"
      }
      return ""
    }
  }
  if {$debug } {  puts "took $loopx loopx loops"}

  # ok, we got the small window... fill the display
  if { [string index [winfo server $w] 0] == "X" } {
  #   wm withdraw $w
   wm iconify $w
    wm attributes $w -fullscreen 1
  } else {
    wm withdraw $w
    wm state $w zoomed
  }
  # Now we wait for the 'fullscreen' to play out...
  while { 1 } {
    set sz "[regsub -all {[x+]}  [winfo geometry $w] { }]"
    if { [lindex $sz 0] == $xs || \
	     [lindex $sz 1] == $ys ||\
	     [lindex $sz 2] == $x ||\
	     [lindex $sz 3] == $y } {
      update
      if { [incr loop2] > 40} { 
	if {$debug } { puts "Loop2 $sz $x $y  $loop2"}
 	return ""
      }
     continue
    }
    # Ok, we now have the display info...
    break
  }
  # Let go of the full screen...
  if { [string index [winfo server $w] 0] == "X" } {
    wm attributes $w -fullscreen 0
  } else {
    wm state $w normal
  }
  # convert to minX maxX minY maxY
  set result "[lindex $sz 2] [expr {[lindex $sz 2] + [lindex $sz 0]} ] \
              [lindex $sz 3] [expr {[lindex $sz 3] + [lindex $sz 1]} ] "
  if {$debug } { puts "$sz $result $x $y $loop $loop2"}
  return  $result
}
#
# This code works with cells in a graph.  The central cell (0,0) is the
# first display.  We set up the grid so we can search all cells around
# the defined (so far) displays..

proc populateGrid { g xs ys } {
  upvar $g gr
  # This function sets ajacent unocupied grid points to "?"
  for { set x [expr {$xs -1}]} {$x <= $xs + 1} {incr x} {
    for { set y [expr {$ys -1}]} {$y <= $ys + 1} {incr y} {
      if { ! [info exists gr($x,$y)] } {
	set gr($x,$y) "?"
      }
    }
  }
}
proc getDisplays { } {
  global debug
  set debug 0

  set result ""
  #
  # Set up screen info for whoever needs to know where we are...
  # Here we figure out how many displays are on this screen
  # and what their boundries are.  Woops, belay that.  I was assuming that
  # the tk definition of a screen was the whole display area and that each
  # 'head' or monitor was a sub part of that.  NOT SO for windows.
  # In windows, the wm screenwidth/screenheight commands return the home
  # display area (the one with the windows start menu).  This appears to be
  # at location "+0+0" and, if another monitor is to the left, it will have
  # negative locations for its orgin and all positions in it.  I assume the
  # same is true of locations above the 'home' monitor.  Also, it is possible
  # for the displays to be offset, even corner to corner.
  set w [toplevel .a ]
  if { [string index [winfo server $w] 0] == "X" } {
 #   wm iconify $w
 
  } else {
    wm overrideredirect $w 1
    wm withdraw $w
  }
  # We depend on fullscreen meaning on the current display, i.e. of the display
  # containing the top left corner of the window.  But then, windows seems to
  # define 'fullscreen' as the home display,fullscreen :(. 

  # start at the beginning...
  set wd [set hd 30]
  set gr(0,0) 0
  set x [set y 0]
  while { 1 } {
    # If this is a new display, add it to the list
    set sz [probeDisplay $w $wd $hd]
    if {$debug} {puts "$x $y $wd $hd ->$sz<"}
    if { $sz != "" } {    
      if { [lsearch -exact $result $sz] == -1 } {
	lappend result $sz
	set gr($x,$y) [expr {[llength $result] - 1} ]
	# Now add the cells around it to the grid
	# This will also reset the 'array nextelement' to start over
	populateGrid gr $x $y
	set sw  [expr {[lindex $sz 1] - [lindex $sz 0]}]
	set sh  [expr {[lindex $sz 3] - [lindex $sz 2]}] 
	set id [array startsearch gr]
      }
    }
      if { ! [info exists id] } {
	if {$debug} {
	  puts "id does not exist, bad thinks follow.  sz =$sz result = $result"
	}
      }
    while { 1 } {
      # here we look at each cell in the grid.  If it is '?' it is possibly
      # a display so we break out, thus redoing the above while...
      if { [set next [array nextelement gr $id]] == "" } {
	destroy .a
	# Ok, now we know how many and a pretty good idea of the sizes, 
	# however, there may be hidden areas such as system trays, etc.
#	puts "$result"
	foreach dis $result {
	  lappend final [findNoMansLand $dis]
	}
	return $final
      }
      set next [split $next "," ]
      set x [lindex $next 0]
      set y [lindex $next 1]
#      puts "$x $y $gr($x,$y) "
      if { $gr($x,$y) == "?" } {
	set gr($x,$y) "n"
	set wd [expr { $x * $sw + $sw /2}]
	set hd [expr { $y * $sh + $sh /2}]
#	puts "wd $x $sw ->$wd | hd  $y $sh -> $hd"
	break
      }
    }
  }
}
###########################################################################
