#!/bin/sh
# the next line restarts using wish \
    exec wish "$0" "$@"

# tcolor --
# This script implements a simple color editor, where you can
# create colors using either the RGB, HSB, or CYM color spaces
# and apply the color to existing applications.
#
# IputColor (start) is the color to start with and to which we return if
# exit is any but OK.
# whichColor(name) is the name of space and the title.
# We create a namespace (named "name").  Each time the color is changed
# we will execute the script "dothis" (unless the autoupdates option is
# disabled - see the menu on the color edit window).  We put up a 
# Color edit window which can be used to change the color in either of the
# three color spaces (RGB HSB & CYM) as per the menu.  The window & 
# color editor are exited by either the OK, Cancel or X buttons which
# also destroy the window and the namespace.  If, per chance, the window
# and name space exist on entry, they will be cleaned up (i.e. destroyed)
# and a new namespace/window is created.

# The window by the name of ".$name" may be distroyed to clean up and 
# destroy the namespace as well.


proc ColorEditor { name dothis start} {
#  puts "dothis is $dothis"
  #Clean up any old calls...
  catch {destroy .$name}

  namespace eval $name {
    variable label1
    variable label1
    variable label2
    variable label3
    variable command
    variable w
    variable red
    variable green
    variable blue
    variable color
    variable colorSpace
    variable updating
    variable autoUpdate
    variable colorOnInput
    variable whichColor [namespace tail [namespace current]]
#    puts "$whichColor"

    proc DoColorEditor {dothis entrycolor} {
      variable label1
      variable label1
      variable label2
      variable label3
      variable command "$dothis"
      variable w
      variable red
      variable green
      variable blue
      variable colorSpace
      variable updating
      variable autoUpdate
      variable colorOnInput
      variable whichColor

 #     puts "do this is $command"


      set colorOnInput $entrycolor
      set color $colorOnInput
      set w .$whichColor
      toplevel $w
      wm title $w "Color Editor $whichColor"
      wm protocol $w WM_DELETE_WINDOW [namespace code {ExitColor 1}]
      #  puts "$color is"
      # variables that control the program:
      #
      # colorSpace -		Color space currently being used for
      #				editing.  Must be "rgb", "cmy", or "hsb".

      # label1, label2, label3 - Labels for the scales.

      # red, green, blue -	Current color intensities in decimal
      #				on a scale of 0-65535.

      # color -			A string giving the current color value
      #				in the proper form for x:
      #				#RRRRGGGGBBBB

      # updating -		Non-zero means that we're in the middle of
      #				updating the scales to load a new color,so
      #				information shouldn't be propagating back
      #				from the scales to other elements of the
      #				program:  this would make an infinite loop.

      # command -		Holds the command that has been typed
      #				into the "Command" entry.

      # autoUpdate -		1 means execute the update command
      #				automatically whenever the color changes.

      # name -			Name for new color, typed into entry.

      set colorSpace hsb
      set red 65535
      set green 0
      set blue 0
      set updating 0
      set autoUpdate 1
      set name ""

      # Create the menu bar at the top of the window.

      frame $w.menu -relief raised -borderwidth 2 
      pack $w.menu -side top -fill x
      menubutton $w.menu.file -text ColorSpace  -menu $w.menu.file.m -underline 0
      menu $w.menu.file.m -tearoff false
      $w.menu.file.m add radio \
	  -label "RGB color space" \
	  -variable [namespace current]::colorSpace \
	  -value rgb \
	  -underline 0 \
	  -command [namespace code {changeColorSpace rgb}]
      $w.menu.file.m add radio \
	  -label "CMY color space" \
	  -variable [namespace current]::colorSpace \
	  -value cmy\
	  -underline 0 \
	  -command [namespace code {changeColorSpace cmy}]
      $w.menu.file.m add radio \
	  -label "HSB color space" \
	  -variable [namespace current]::colorSpace \
	  -value hsb \
	  -underline 0 \
	  -command [namespace code {changeColorSpace hsb}]
      $w.menu.file.m add separator
      $w.menu.file.m add radio\
	  -label "Automatic updates"\
	  -variable [namespace current]::autoUpdate \
	  -command [namespace code {doUpdate}] \
	  -value 1 \
	  -underline 0
      $w.menu.file.m add radio \
	  -label "Manual updates" \
	  -variable [namespace current]::autoUpdate \
	  -value 0 \
	  -underline 0
      $w.menu.file.m add separator
      $w.menu.file.m add command \
	  -label "Exit program" \
	  -underline 0 \
	  -command [namespace code {ExitColor  1}]
      pack $w.menu.file -side left

      # Create the command entry window at the bottom of the window, along
      # with the update button.

      frame $w.bot -relief raised -borderwidth 2
      pack $w.bot -side bottom -fill x
      #label $w.commandLabel -text "Command:"
      #entry $w.command -relief sunken -borderwidth 2 -textvariable command \
	  #	-font -Adobe-Courier-Medium-R-Normal--*-120-*-*-*-*-*-*
      button $w.ok -text OK -command [namespace code {ExitColor 0}]
      button $w.cancel -text Cancel -command [namespace code {ExitColor 1}]
      #pack $w.commandLabel -in $w.bot -side left
      pack $w.cancel -in $w.bot -side right -pady .1c -padx .25c
      pack $w.ok -in $w.bot -side right -pady .1c -padx .25c
      #pack $w.command -in $w.bot -expand yes -fill x -ipadx 0.25c

      # Create the listbox that holds all of the color names in rgb.txt,
      # if an rgb.txt file can be found.

      frame $w.middle -relief raised -borderwidth 2
      pack $w.middle -side top -fill both
      foreach i {
          /usr/share/X11/rgb.txt
	  /usr/local/lib/X11/rgb.txt /usr/lib/X11/rgb.txt
	  /X11/R5/lib/X11/rgb.txt /X11/R4/lib/rgb/rgb.txt
	  /usr/openwin/lib/X11/rgb.txt} {
	if ![file readable $i] {
	  continue;
	}
	set f [open $i]
	frame $w.middle.left
	pack $w.middle.left -side left -padx .25c -pady .25c
	listbox $w.names -width 20 -height 12 -yscrollcommand "$w.scroll set" \
	    -relief sunken -borderwidth 2 -exportselection false
	# The following regsub dance extracts the window name (which is also
	# the space name) from %W. 
	bind $w.names <1> \
	    [namespace code {tc_loadNamedColor \
	    [%W get [%W nearest %y]]}]

	scrollbar $w.scroll -orient vertical -command "$w.names yview" \
	    -relief sunken -borderwidth 2
	pack $w.names -in $w.middle.left -side left
	pack $w.scroll -in $w.middle.left -side right -fill y
	while {[gets $f line] >= 0} {
	    regsub -all {\"[^ ]} $line {\" } line ;
	    if {[llength $line] == 4} {
		    if {[string first {#} $line] == -1 } {
			    $w.names insert end [lindex $line 3]
			}
		}
	}
	close $f
	break
      }

      # Create the three scales for editing the color, and the entry for
      # typing in a color value.

      frame $w.middle.middle
      pack $w.middle.middle -side left -expand yes -fill y
      frame $w.middle.middle.1
      frame $w.middle.middle.2
      frame $w.middle.middle.3
      frame $w.middle.middle.4
      pack $w.middle.middle.1 $w.middle.middle.2 $w.middle.middle.3 \
	  -side top -expand yes
      pack $w.middle.middle.4 -side top -expand yes -fill x
      foreach i {1 2 3} {
	label $w.label$i -textvariable [namespace current]::label$i
	scale $w.scale$i -from 0 -to 1000 -length 6c -orient horizontal \
	    -command [namespace code {tc_scaleChanged}]
	pack $w.scale$i $w.label$i -in $w.middle.middle.$i -side top -anchor w
      }
      #label $w.nameLabel -text "Name:"
      #entry $w.name -relief sunken -borderwidth 2 -textvariable name -width 10 \
	  #	-font -Adobe-Courier-Medium-R-Normal--*-120-*-*-*-*-*-*
      #pack $w.nameLabel -in $w.middle.middle.4 -side left
      #pack $w.name -in $w.middle.middle.4 -side right -expand 1 -fill x
      #bind $w.name <Return> {tc_loadNamedColor $name color}

      # Create the color display swatch on the right side of the window.

      frame $w.middle.right
      pack $w.middle.right -side left -pady .25c -padx .25c -anchor s
      frame $w.swatch -width 2c -height 5c -background $color
      label $w.value -textvariable [namespace current]::color -width 13 \
	  -font -Adobe-Courier-Medium-R-Normal--*-120-*-*-*-*-*-*
      pack $w.swatch -in $w.middle.right -side top -expand yes -fill both
      pack $w.value -in $w.middle.right -side bottom -pady .25c

      tc_loadNamedColor $colorOnInput

      changeColorSpace hsb
#      puts "return ce $whichColor"
    }
    # The procedure below is invoked when one of the scales is adjusted.
    # It propagates color information from the current scale readings
    # to everywhere else that it is used.

    proc tc_scaleChanged { args} {
      variable red
      variable green
      variable blue
      variable color
      variable colorSpace
      variable updating
      variable autoUpdate
      variable whichColor
      variable w
      if $updating {
	return
      }
      if {$colorSpace == "rgb"} {
	set red   [format %.0f [expr [$w.scale1 get]*65.535]]
	set green [format %.0f [expr [$w.scale2 get]*65.535]]
	set blue  [format %.0f [expr [$w.scale3 get]*65.535]]
      } else {
	if {$colorSpace == "cmy"} {
	  set red   [format %.0f [expr {65535 - [$w.scale1 get]*65.535}]]
	  set green [format %.0f [expr {65535 - [$w.scale2 get]*65.535}]]
	  set blue  [format %.0f [expr {65535 - [$w.scale3 get]*65.535}]]
	} else {
	  set list [hsbToRgb [expr {[$w.scale1 get]/1000.0}] \
			[expr {[$w.scale2 get]/1000.0}] \
			[expr {[$w.scale3 get]/1000.0}]]
	  set red [lindex $list 0]
	  set green [lindex $list 1]
	  set blue [lindex $list 2]
	}
      }

      set color [format "#%04x%04x%04x" $red $green $blue]
      $w.swatch config -bg $color
#      puts "$autoUpdate autoupdate"
      if {$autoUpdate != 0 } {
	doUpdate
      }
      update idletasks
    }

    # The procedure below is invoked to update the scales from the
    # current red, green, and blue intensities.  It's invoked after
    # a change in the color space and after a named color value has
    # been loaded.

    proc tc_setScales {} {
      variable red
      variable green
      variable blue
      variable colorSpace
      variable updating
      variable w
 #     puts "setting scales $colorSpace for $w"
      set updating 1
      if {$colorSpace == "rgb"} {
	$w.scale1 set [format %.0f [expr $red/65.535]]
	$w.scale2 set [format %.0f [expr $green/65.535]]
	$w.scale3 set [format %.0f [expr $blue/65.535]]
      } else {
	if {$colorSpace == "cmy"} {
	  $w.scale1 set [format %.0f [expr (65535-$red)/65.535]]
	  $w.scale2 set [format %.0f [expr (65535-$green)/65.535]]
	  $w.scale3 set [format %.0f [expr (65535-$blue)/65.535]]
	} else {
	  set list [rgbToHsv $red $green $blue]
	  $w.scale1 set [format %.0f [expr {[lindex $list 0] * 1000.0}]]
	  $w.scale2 set [format %.0f [expr {[lindex $list 1] * 1000.0}]]
	  $w.scale3 set [format %.0f [expr {[lindex $list 2] * 1000.0}]]
	}
      }
      set updating 0
    }

    # The procedure below is invoked when a named color has been
    # selected from the listbox or typed into the entry.  It loads
    # the color into the editor.

    proc tc_loadNamedColor {name } {    

      variable red
      variable green
      variable blue
      variable color
      variable  autoUpdate
      variable w
      if {[string index $name 0] != "#"} {
	set list [winfo rgb $w.swatch $name]
#      puts "$name $list"
	set red [lindex $list 0]
	set green [lindex $list 1]
	set blue [lindex $list 2]
#      puts "1red $red green $green blue $blue"
      } else {
	case [string length $name] {
	  4 {set format "#%1x%1x%1x"; set shift 12}
	  7 {set format "#%2x%2x%2x"; set shift 8}
	  10 {set format "#%3x%3x%3x"; set shift 4}
	  13 {set format "#%4x%4x%4x"; set shift 0}
	  default {error "syntax error in color name \"$name\""}
	}
	if {[scan $name $format red green blue] != 3} {
	  error "syntax error in color name \"$name\""
	}
	set red [expr $red<<$shift]
	set green [expr $green<<$shift]
	set blue [expr $blue<<$shift]
      }
      tc_setScales
#      puts "2red $red green $green blue $blue"
      set color [format "#%04x%04x%04x" $red $green $blue]
      $w.swatch config -bg $color
      if {$autoUpdate != 0 } {
	doUpdate
      }
    }

    # The procedure below is invoked when a new color space is selected.
    # It changes the labels on the scales and re-loads the scales with
    # the appropriate values for the current color in the new color space

    proc changeColorSpace space {
      variable label1
      variable label2
      variable label3
      variable w
#      puts "setting space lables $space"
      if {$space == "rgb"} {
	set label1 Red
	set label2 Green
	set label3 Blue
	tc_setScales
	return
      }
      if {$space == "cmy"} {
	set label1 Cyan
	set label2 Magenta
	set label3 Yellow
	tc_setScales
	return
      }
      if {$space == "hsb"} {
	set label1 Hue
	set label2 Saturation
	set label3 Brightness
	tc_setScales
	return
      }
    }

    # The procedure below converts an RGB value to HSB.  It takes red, green,
    # and blue components (0-65535) as arguments, and returns a list containing
    # HSB components (floating-point, 0-1) as result.  The code here is a copy
    # of the code on page 615 of "Fundamentals of Interactive Computer Graphics"
    # by Foley and Van Dam.

    proc rgbToHsv {red green blue} {
      if {$red > $green} {
	set max $red.0
	set min $green.0
      } else {
	set max $green.0
	set min $red.0
      }
      if {$blue > $max} {
	set max $blue.0
      } else {
	if {$blue < $min} {
	  set min $blue.0
	}
      }
      set range [expr $max-$min]
      if {$max == 0} {
	set sat 0
      } else {
	set sat [expr {($max-$min)/$max}]
      }
      if {$sat == 0} {
	set hue 0
      } else {
	set rc [expr {($max - $red)/$range}]
	set gc [expr {($max - $green)/$range}]
	set bc [expr {($max - $blue)/$range}]
	if {$red == $max} {
	  set hue [expr {.166667*($bc - $gc)}]
	} else {
	  if {$green == $max} {
	    set hue [expr {.166667*(2 + $rc - $bc)}]
	  } else {
	    set hue [expr {.166667*(4 + $gc - $rc)}]
	  }
	}
	if {$hue < 0.0} {
	  set hue [expr $hue + 1.0]
	}
      }
      return [list $hue $sat [expr {$max/65535}]]
    }

    # The procedure below converts an HSB value to RGB. It takes hue, saturation,
    # and value components (floating-point, 0-1.0) as arguments, and returns a
    # list containing RGB components (integers, 0-65535) as result.  The code
    # here is a copy of the code on page 616 of "Fundamentals of Interactive
    # Computer Graphics" by Foley and Van Dam.

    proc hsbToRgb {hue sat value} {
      set v [format %.0f [expr 65535.0*$value]]
      if {$sat == 0} {
	return "$v $v $v"
      } else {
	set hue [expr $hue*6.0]
	if {$hue >= 6.0} {
	  set hue 0.0
	}
	scan $hue. %d i
	set f [expr $hue-$i]
	set p [format %.0f [expr {65535.0*$value*(1 - $sat)}]]
	set q [format %.0f [expr {65535.0*$value*(1 - ($sat*$f))}]]
	set t [format %.0f [expr {65535.0*$value*(1 - ($sat*(1 - $f)))}]]
	case $i \
	    0 {return "$v $t $p"} \
	    1 {return "$q $v $p"} \
	    2 {return "$p $v $t"} \
	    3 {return "$p $q $v"} \
	    4 {return "$t $p $v"} \
	    5 {return "$v $p $q"}
	error "i value $i is out of range"
      }
    }

    # The procedure below is invoked when the "Update" button is pressed,
    # and whenever the color changes if update mode is enabled.  It
    # propagates color information as determined by the command in the
    # Command entry.

    proc doUpdate {} {
      variable color
      variable  command
      variable autoUpdate
      set newCmd $command
      regsub -all %% $command $color newCmd
      eval $newCmd
    }
    proc ExitColor { zip} {
      variable color
      variable w
      variable whichColor
      variable colorOnInput

      if { $zip == 1 } {
	set color $colorOnInput
      }
      doUpdate
      catch "destroy  $w"
      balloonhelp_cancel
      catch "namespace delete ::$whichColor"
#      puts "exiting ce $whichColor"
    }
  }
  namespace eval $name "DoColorEditor {$dothis} $start"
  return .$name
}
#changeColorSpace hsb
#if {[lindex $argv 0] != ""} {

#  catch {tc_loadNamedColor [lindex $argv 0] color}
#}

