# -*- tcl -*- \
    last modified <20120318.1640.29>

proc GetTimeFromSecs { time } {
  global dateformat
  # Convert time in sec (UTC) to either yymmdd hh:mm:ss or ddmmyy hh:mm:ss
  return [clock format $time -format $dateformat]
}
proc GetTimeFromSecsSetFormat { format } {
  global dateformat
  # 0 -> yymmdd
  # 1 -> ddmmyy
  if { $format == 0 } {
    set dateformat "%y%m%d %H:%M:%S"
  } elseif { $format == 1 } {
    set dateformat "%d%m%y %H:%M:%S"
  } else {
    set dateformat $format
  }
}

 
proc ClockMilliSeconds { } {
  return [clock milliseconds]
}

proc GetEuid { } {
  global glob
  
  if { $glob(os) == "WIN32" } {
    return "1"
  }
  return "whodat"
}

proc GetUidGidString { file uid gid } {
  global glob users groups
  
  if { $glob(os) == "WIN32" } {
    return ""
  }
  if { ![info exists users($uid)] || ![info exists groups($gid)] } {
    # Need to work for it...
    if { [catch {file attributes $file} gu] != 0 } {
      # rare but happens with dead links...
      return ""
    }
    set gu [file attributes $file]
    set users($uid) [lindex $gu 3]
    set groups($gid) [lindex $gu 1]
  }
  return "$users($uid)/$groups($gid)" 
}

proc GetDF { dir } {
  # Makes a stab at getting the full size of the disk
  global glob
  set r [catch "exec df {$dir}" out]
  set oout [split $out]
  set indx [lsearch -regexp $oout {[0-9]%}]
  if { $indx != -1 } {
    incr indx -1
    while {[set free [lrange $oout $indx $indx]] == "{}" } {
      incr indx -1
    }
#    puts "$oout"
#    puts "$free"
    set indx 1

  } elseif {$glob(os) == "WIN32" &&
	    [catch {exec cmd /C dir} out] == 0 } {
    set oout [split $out "\n"]
    set free [split [lrange $oout end end]]
    set indx [lsearch -exact $free bytes]
    incr indx -1
    set free [lrange $free $indx $indx]
    set indx 0
  } else {
    LogSilent $out
    return "??"
  }
  regsub -all {[^0-9]} $free "" free
  # should now be a pure number.  If nil, use zero...
  # better than failing in the following code...
  if { $free == "" } {
    set free 0
  }
  while { $free > 1024. } {
    incr indx
    set free [expr {$free / 1024.}]
  }
#  puts "$indx $free"
  set suf { " " K M G T}
  return [format {%.3f%s} $free [lrange $suf $indx $indx]]
}
